class Standarts:
    '''
    # Standarts
    Get standart functions and code templates.
    '''
    _ = ''
    def simpleparser(self) -> str:
        '''
        get code of bytex2 parser and splitter
        '''
        code = """
def parse(code: str) -> list[list[str]]:
    '''
    # Parser
    Simple Bytex2 code splitter and parser.
    '''
    code = code.replace(';', '\n')
    commands = []
    for i in code.split('\n'):
        command = []
        i = i.strip()
        for j in i.split(' '):
            command.append(j)
        if command != ['']:
            commands.append(command)
    return commands
        """
        return code


from UnitVerge import *



class Generator:
    def __init__(self) -> None:
        self.unit = Unit()
        self.unit.new_point('pre')
        self.unit.new_point('post')
        self.unit.pipeline(['pre', 'main', 'post'])
        self.btx = Bytex().translator()
    
    def goto(self, name):
        self.unit.set_point(name)
    
    def __call__(self, line: str) -> Unit:
        return self.unit(line)
    
    def py(self, line: str):
        self.unit.python(line)
    
    def build(self):
        verge = Verge()
        verge.interprete([self.unit])
        return self.btx.translate('\n'.join(verge.compile()))


from .syntax import Syntax
import re







class Lexer:
    def __init__(self, code):
        self.code = code
        self.pos = 0
        self.tokens = []
    
    def tokenize(self):
        while self.pos < len(self.code):
            matched = False
            if re.match(r'\s', self.code[self.pos]):
                self.pos += 1
                continue
            for token_type, pattern in Syntax.syntax.items():
                regex = re.compile(pattern)
                match = regex.match(self.code, self.pos)
                if match:
                    if token_type not in ['COMMENT', 'NONE']:
                        self.tokens.append({
                            'type': token_type,
                            'value': match.group(),
                            'position': self.pos
                        })
                    self.pos = match.end()
                    matched = True
                    break
            if not matched:
                raise SyntaxError(f"Unknown syntax: {self.code[self.pos]}")
        
        return self.tokens
    
    
    

code = '''
int test = 10
'''


print(Lexer(code).tokenize())


'''
# BYTEX2
Main file of bytex translator.

## Extend:
    parse(code: str) -> list[list[str]]

        : Simple language parser and splitter.
    
    Generator : metaclass

        : Metaclass for code generation.
    
    Translator : class

        : Main class for translation bytex to python instructions.
'''



from .exceptions import *
from .basics import core



def parse(code: str) -> list[list[str]]:
    '''
    # Parser
    Simple Bytex2 code splitter and parser.
    '''
    code = code.replace(';', '\n')
    commands = []
    for i in code.split('\n'):
        command = []
        i = i.strip()
        for j in i.split(' '):
            command.append(j)
        if command != ['']:
            commands.append(command)
    return commands

















# metaclass
class Generator:
    '''
    # Generator
    Metaclass for code generation.
    '''
    def __init__(self):
        self._code = []
        self._pre = []
        self._end = []
        self._log = []
        self.tabs = 0
    
    def _line(self, code: str):
        '''
        Add code (to _code)
        '''
        self._code.append('    ' * self.tabs + code)
    
    @property
    def code(self):
        '''
        Returning all builded code in format:

        - core code
        - raw code   (_log)
        - pre code   (_pre)
        - main code  (_code)
        - pos code   (_end)
        '''
        code = f'''
# =============================
#         core code
# =============================
{'\n'.join(core)}



# =============================
#          raw code
# =============================
RAW = """
{'\n'.join(self._log)}
"""


# =============================
#      compiled code: pre
# =============================
{'\n'.join(self._pre)}


# =============================
#     compiled code: main
# =============================
{'\n'.join(self._code)}



# =============================
#     compiled code: post
# =============================
{'\n'.join(self._end)}
        '''
        return code.split('\n')











class Translator(Generator):
    def __init__(self, safemode: bool = False):
        super().__init__()
        self.in_struct = 0
        self.included = []
        self.user_structs = {
            'COMMAND':{'PARSER': lambda context, line: 0 } # EXAMPLE
            # context - [Translator] object
            # line - full line ([0] - is command)
        }
        self.writepython = False
        self.safe = safemode
        
    def translate(self, code: str):
        code = parse(code)
        self.parse(code)
        return self.code
    
    def command(self, command: str, parser: object):
        '''
        Add command to language.
        
        Args:
            - command : str 
                - command to parse
            
            - parser : object - function (context: Translator, line: list[str])
                - object to parse command
        '''
        self.user_structs[command] = {'PARSER': parser}
        return self
    
    def pre_edit(self, code: list[str]):
        for line in code: 
            self._pre.append(line)
        return self

    def post_edit(self, code: list[str]):
        for line in code: 
            self._end.append(line)
        return self

    def main_edit(self, code: list[str]):
        for line in code: 
            self._code.append(line)
        return self
    
    def plugin(self, type: str = 'command'):
        '''
        Add Plugin to language. 
        
        types: 
            - command 
                -> function (command: str, parser: object)
                : add command to language
            - main
                -> function (code: list[str])
                : edit main code
            - pre 
                -> function (code: list[str])
                : edit code preprocess
            - post
                -> function (code: list[str])
                : edit finalization code
        '''
        if type == 'command':
            return self.command
        elif type == 'pre':
            return self.pre_edit
        elif type == 'post':
            return self.post_edit
        elif type == 'main':
            return self.main_edit
        else:
            raise Bytex2Error('Plugin error', f'Unknown type of plugin: {type}')
        
    def parse(self, code: list):
        for line in code:
            try: 
                self._log.append(
                    str(code.index(line)) + "- " +
                    ' '.join(line)
                )
                if self.writepython and line[0] != 'python': 
                    if not self.safe:
                        self._line(" ".join(line).replace('\\tab|', '    '))
                    else:
                        raise RuntimeError('Safemode error', 'Turn off safe mode.')
                else:
                    self.work(line)
            except Exception as e: 
                raise TranslationError('Interpritation error', str(e), line, code, 1)
                
    
    
    
    def work(self, line: list[str]):
        command = line[0]
        self._line(f'# line: {repr(" ".join(line))}')
        if command.startswith('#') and command != '#':
            self._preprocess(line)
            return
        elif command.startswith(('$', '@')):
            self._system(line)
            return
        elif command.startswith('!'):
            self._param(line)
            return
        elif command in self.user_structs.keys():
            try: self.user_structs[command]['PARSER'](self, line)
            except Exception as e:
                raise Bytex2Error(f'Plugin exception', str(e))
        else:
            self._main(line)
            return
    
    
    
    
    
    def _preprocess(self, line):
        command = line[0]
        args = line[1:]
        def checklen(length: int = 0, operator: str = "!="):
            if not eval(f'len(args) {operator} length'):
                raise SyntaxError(f"Invalid argument length: {len(args)} (must be {length})")
        match command:
            case '#append':
                checklen(1, '==')
                try:
                    with open(f"{args[0]}.btx", 'r') as f:
                        content = f.read()
                except FileNotFoundError:
                    raise Bytex2Error('Include error', 
                            f"Module {name} is not found.")
                compiled = Translator().translate(content)
                self._pre.append(f'# APPEND: {args[0]}')
                self._pre.append(f'class {args[0].split("/")[-1]}:')
                [self._pre.append('    ' + str(i))
                for i in compiled]
            
            case '#include':
                checklen(1, '>=')
                if self.safe:
                    raise Bytex2Error("Include error', 'Include is not available in safemode. Turn off safemode or use '#append'.")
                name = args[0]
                libs = [name] if len(args) == 1 else args[1:]
                append = f'{name}{libs}'
                if append in self.included: return
                else: 
                    self.included.append(append)
                try:
                    with open(f"{name}.btx", 'r') as f:
                        content = f.read()
                except FileNotFoundError:
                    raise Bytex2Error('Include error', 
                            f"Module {name} is not found.")
                localscope = locals()
                code = Translator().translate(content)
                code = '\n'.join(code)
                exec(code, localscope, localscope)
                if localscope['_is_lib']:
                    if localscope['in_lib']:
                        raise Bytex2Error('Include error', 
                            "Lib is not finished or unknown syntax. Did you forget add '#ELIB' block?")
                    for lib in libs:
                        try:self.pre_edit([f'{lib} = """{localscope[f"LIB{lib}"]}"""'])
                        except KeyError:
                            raise Bytex2Error('Include error', 
                                f"Lib {lib} is not found in file.")
                        self.pre_edit([f'exec({lib})'])
                        self.pre_edit([f'libs.append("{lib}")'])
                        self.work(['create', 'lib_' + lib, lib])
                else:
                    raise Bytex2Error('Include error', "File has no lib. Did you mean '#append'?")
           
            case '#import':
                checklen(1, '>=')
                libs = args[0:]
                for lib in libs:
                    try:
                        __import__(str(lib))
                    except ImportError:
                        raise Bytex2Error('Import error', f"Lib {lib} is not found.")
                    self._pre.append(f'import {lib}')
            
            case '#SLIB': # start LIB
                checklen(1, '==')
                self._line('in_lib = True')
                self._line(f'LIB{args[0]} = """\n')
                self.work(['struct', "lib_" + args[0]])
            
            case '#ELIB': # end LIB
                checklen(0, '==')
                self.work(['endstruct'])
                self._line('"""; in_lib = False')
            
            case _:
                raise SyntaxError('Unknown preprocess command')
    
    
    
    
    
    def _system(self, line):
        command = line[0]
        args = line[1:]
        def checklen(length: int = 0, operator: str = "!="):
            if not eval(f'len(args) {operator} length'):
                raise SyntaxError(f"Invalid argument length: {len(args)} (must be {length})")
        match command:
            case '@islib':
                checklen(0, '==')
                self._pre.append(f'_is_lib = True')
                
            case '$jit':
                self._line(f'@jit({" ".join(args)})')
            
            case '$check':
                arg = args[0]
                match arg:
                    case 'unsafe':
                        if not self.safe:
                            self._line('raise SystemError("\nCode execution must be in safemode.")')
                    case 'safe':
                        if self.safe:
                            self._line('raise SystemError("\nCode execution must be without safemode.")')
                    case 'lib':
                        self.main_edit([
                            'if not _is_lib:'
                            '    raise SystemError("\nCode must be lib.")'
                        ])
                    case _:
                        cond = " ".join(args[0:])
                        self._line(f'if {cond}:')
                        self._line(f'    raise SystemError("\\nSystem check failed ({cond}).")') 
            
            case '$require':
                self._line(f'if not LANG.haslib("{args[0]}"):')
                self._line(f'''    raise ImportError("Lib {args[0]} is not included. Did you forgot wtite '#include [module] {args[0]}'?")''')
            
            case '$addtabs':
                self.tabs += 1

            case '$subtabs':
                self.tabs -= 1
            
            case _:
                raise SyntaxError('Unknown system command')
    
    
    
    
    def _param(self, line):
        command = line[0]
        args = line[1:]
        def checklen(length: int = 0, operator: str = "!="):
            if not eval(f'len(args) {operator} length'):
                raise SyntaxError(f"Invalid argument length: {len(args)} (must be {length})")
        match command:
            case '!ldtp': # load data to parameter
                checklen(2, '==')
                addr = args[0]
                self._line(f'MEM.setto_p({args[1]}, {addr})')
            
            case '!ldfp': # load data from parameter
                checklen(1, '==')
                self._line(f'MEM.setdata(\n{"    " * (self.tabs + 1)}MEM.getfrom_p({args[0]})\n{"    " * self.tabs})')
            
            case '!gdfp': # get data from parameter
                checklen(2, '==')
                addr = args[0]
                self._line(f'{args[1]} = MEM.getfrom_p({addr})')
            
            case _:
                raise SyntaxError('Unknown parameter command')
    
    
    
    def _main(self, line):
        command = line[0]
        args = line[1:]
        def operation(a, b, op):
            self._line(f'# operation {a} {op} {b}')
            self._line(f'__res = MEM.getdata({a}) {op} MEM.getdata({b})')
            self._line(f'MEM.setdatato({a}, __res)')
        def operationreg(rega, regb, a, b, op):
            self._line(f'# register operation')
            self._line(f'# {rega}:{a} {op} {regb}:{b}')
            self._line(f'__res = MEM.getreg({rega})[{a}] {op} MEM.getreg({regb})[{b}]')
            self._line(f'MEM.regs[{rega}][{a}] = __res')
        def checklen(length: int = 0, operator: str = "!="):
            if not eval(f'len(args) {operator} length'):
                raise SyntaxError(f"Invalid argument length: {len(args)} (must be {length})")
                
        match command:
            case 'start': # declarate point
                self._line(f'# start of point {args[0]}')
                self._pre.append(f'points.append("{args[0]}")')
                self._line(f'def {args[0]}():')
                self.tabs += 1
                self._line('pass')
            
            case 'python':
                self.writepython = not self.writepython
            
            case 'func': # declarate function
                checklen(1, ">=")
                self._line(f'# start of function {args[0]}')
                if self.in_struct:
                    func_args = "self, " + " ".join(args[1:])
                    self._line(f'def {args[0]}({func_args}):')
                    self.tabs += 1
                    self._line('pass')
                else:
                    raise SyntaxError('Must be in struct')
        
            case 'sfunc': # declarate static function
                checklen(1, ">=")
                self._line(f'# start of function {args[0]}')
                func_args = " ".join(args[1:])
                self._line("@staticmethod")
                self._line(f'def {args[0]}({func_args}):')
                self.tabs += 1
                self._line('pass')
            
            case 'stdfunc':
                checklen(1, ">=")
                self._line(f'# start of function {args[0]}')
                func_args = " ".join(args[1:])
                self._line(f'def {args[0]}({func_args}):')
                self.tabs += 1
                self._line('pass')
            
            case 'struct':
                checklen(1, "==")
                self._line(f'# start of structure {args[0]}')
                self._line(f'class {args[0]}:')
                self.in_struct += 1
                self.tabs += 1
                self._line('_ = None')

            case 'VM':
                checklen(1, ">=")
                self._line(f'MEM.{args[0]} {" ".join(args[1:])}')
            
            case 'self':
                checklen(1, ">=")
                if self.in_struct:
                    self._line(f'self.{args[0]} {" ".join(args[1:])}')
                else:
                    raise SyntaxError('Must be in struct')
            
            case 'create':
                checklen(2, ">=")
                self._line(f'# creating object')
                class_name = args[0]
                obj_name = args[1]
                constr_args = args[2:] if len(args) > 2 else []
                self._line(f'{obj_name} = {class_name}({" ".join(constr_args)})')
            
            case 'do':
                checklen(1, ">=")
                func_name = args[0]
                func_args = args[1:] if len(args) > 1 else []
                self._line(f'{func_name}({" ".join(func_args)})')

            case 'loop':
                checklen(0, "==")
                self._line(f'# loop')
                self._line(f'while True:')
                self.tabs += 1
                self._line('None')

            case 'repeat':
                checklen(1, "==")
                self._line(f'# repeat loop')
                self._line(f'for i in range({args[0]}):')
                self.tabs += 1
                self._line('None')
            
            case 'iters':
                checklen(2, "==")
                self._line(f"if {repr(args[0])} in points: MEM.repeat({args[0]}, {args[1]})")
            
            case 'delay':
                checklen(1, "==")
                self._line(f'time.sleep({args[0]} * 0.001)')
                
            case 'break':
                self._line(f'# break loop')
                if len(args) > 0:
                    if args[0] == 'if':
                        operator = args[1]
                        value = args[2]
                        self._line(f'if MEM.getdata() {operator} {value}:break')
                else: self._line('break')
            
            case 'end': # end of block
                checklen(0, "==")
                if self.tabs >= 1: self.tabs -= 1
            
            case 'endstruct': # end of structure
                checklen(0, "==")
                if self.tabs >= 1: 
                    self.tabs -= 1; 
                    self.in_struct -= 1
                
            case 'try':
                checklen(0, "==")
                self._line('try:')
                self.tabs += 1
                self._line('None')

            case 'err':
                checklen(0, "==")
                self._line('except Exception as e:')
                self.tabs += 1
                self._line('None')
            
            case 'handle':
                checklen(0, "==")
                self._line('except Exception as e: print(e)')
                
            case 'debug':
                checklen(0, "==")
                self._line('print(f"DEBUG: CURSOR={MEM.CURSOR}, ' \
                    'POINTER={MEM.POINTER}, DATA={MEM.getdata()}, HAND={MEM.hand}")')
                
            case 'goto':
                checklen(1, "==")
                self._line(f'if "{args[0]}" in points:')
                self.tabs += 1
                self._line(f'{args[0]}()')
                self.tabs -= 1
                self._line(f'else:')
                self.tabs += 1
                self._line(f'raise EXECUTIONERROR("Unknown point: {args[0]}")')
                self.tabs -= 1
            
            case 'save':
                checklen(1, "==")
                name = args[0]
                self._line(f'with open({name}, "wb") as f:')
                self._line(f'    pickle.dump(MEM, f)')
                
            case 'open':
                checklen(1, "==")
                name = args[0]
                self._line(f'with open({name}, "rb") as f:')
                self._line(f'    content = pickle.load(f)')
                self._line(f'MEM = content')
                
            case 'copy':
                checklen(1, "==")
                spl = args[0].split(':')
                self._line(f'MEM.setdataraw({spl[0]}, {spl[1]}, MEM.getdata())')
                
            case 'move':
                checklen(2, "==")
                self._line(f'MEM.setdatato({args[0]}, MEM.getdata())')
            
            case 'moveto':
                checklen(1, "==")
                self._line(f'MEM.CURSOR = int({args[0]})')
                
            case 'spawn':
                checklen(1, "==")
                point = args[0]
                self._line(f'threading.Thread(target={point}).start()')
                
            case 'error':
                checklen(1, "==")
                name = f'Error{args[0]}'
                self._line(f'class {name}(Exception):pass')
                self._line(f'raise {name}')
                
            case 'jump':
                checklen(1, "==")
                self._line(f'MEM.POINTER = int({args[0]})')
                
            case 'load':
                checklen(1, "==")
                self._line(f'MEM.setdata({args[0]})')
                
            case 'src':
                checklen(1, ">=")
                if not self.safe:
                    self._line(' '.join(args))
                else:
                    raise SyntaxError \
                (f"Unknown syntax: System commands is not available in safemode")
            
            case '>':
                checklen(0, "==")
                self._line('__p = MEM.CURSOR')
                self._line('MEM.CURSOR = __p + 1')
            
            case '<':
                checklen(0, "==")
                self._line('__p = MEM.CURSOR')
                self._line('MEM.CURSOR = __p - 1')
            
            case '+':
                checklen(0, "==")
                self._line(f'__a = MEM.getdata()')
                self._line(f'MEM.setdatato(MEM.POINTER, __a + 1)')
            
            case '-':
                checklen(0, "==")
                self._line(f'__a = MEM.getdata()')
                self._line(f'MEM.setdatato(MEM.POINTER, __a - 1)')
            
            case 'in':
                checklen(0, "==")
                self._line('MEM.setdata(int(input()))')
            
            case 'sum':
                checklen(2, "==")
                reg = args[0]
                name = args[1]
                self._line(f'{name} = MEM.getsum({reg})')
            
            case 'var':
                checklen(2, ">=")
                self._line(f'{args[0]} = {" ".join(args[1:])}')
                
            case 'add':
                checklen(2, "==")
                a = args[0]
                b = args[1]
                operation(a, b, '+')
            
            case 'sub':
                checklen(2, "==")
                a = args[0]
                b = args[1]
                operation(a, b, '-')
                
            case 'mul':
                checklen(2, "==")
                a = args[0]
                b = args[1]
                operation(a, b, '*')
                
            case 'div':
                checklen(2, "==")
                a = args[0]
                b = args[1]
                operation(a, b, '/')
                
            case 'pow':
                checklen(2, "==")
                a = args[0]
                b = args[1]
                operation(a, b, '**')
                
            case 'addreg':
                checklen(2, "==")
                a_spl = args[0].split(':')
                b_spl = args[1].split(':')
                operationreg(a_spl[0], b_spl[0], a_spl[1], b_spl[1], '+')
            
            case 'subreg':
                checklen(2, "==")
                a_spl = args[0].split(':')
                b_spl = args[1].split(':')
                operationreg(a_spl[0], b_spl[0], a_spl[1], b_spl[1], '-')
                
            case 'divreg':
                checklen(2, "==")
                a_spl = args[0].split(':')
                b_spl = args[1].split(':')
                operationreg(a_spl[0], b_spl[0], a_spl[1], b_spl[1], '/')
                
            case 'mullreg':
                checklen(2, "==")
                a_spl = args[0].split(':')
                b_spl = args[1].split(':')
                operationreg(a_spl[0], b_spl[0], a_spl[1], b_spl[1], '*')
                
            case 'powreg':
                checklen(2, "==")
                a_spl = args[0].split(':')
                b_spl = args[1].split(':')
                operationreg(a_spl[0], b_spl[0], a_spl[1], b_spl[1], '**')
                
            case 'switch':
                checklen(1, "==")
                self._line(f'MEM.sethand("{args[0]}")')
                
            case 'new':
                checklen(1, "==")
                self._line(f'MEM.newhand("{args[0]}")')
                
            case 'reg':
                checklen(1, "==")
                self._line(f'MEM._create_reg({args[0]})')
                
            case 'workif':
                checklen(3, "==")
                operator = args[0]
                value = args[1]
                call = args[2]
                self._line(f'if MEM.getdata() {operator} {value}:{call}()')
                
            case 'out.':
                checklen(0, "==")
                self._line('print(MEM.getdata(), end = "")')
                
            case 'outch':
                checklen(0, "==")
                self._line('print(chr(MEM.getdata()), end = "")')
                
            case 'echo':
                self._line(f'print({" ".join(args)}, end = "")')
            
            case 'echoln':
                self._line(f'print({" ".join(args)})')
                
            case '#': self._line(f'# {" ".join(args)}')
            
            case _:
                raise SyntaxError(f"Unknown syntax: {' '.join(line)}")


import re
import sys
from typing import Any
from .exceptions import PipelineError
from .logger import Logger


class instruction:
    '''
    # Framework instruction metaclass 
    For your Verge-instructions. Only system methods for Verge.
    '''
    def __init__(self, name: str) -> None:
        self.name = str(name)
    def __repr__(self):
        return f'instruction-{self.name}'







class generatable(instruction):
    '''
    # Framework generate metaclass-instruction
    Metclass for code-generators.

    Methods:

    - new_point (name: str) -> None

        : Method for create code point

    - set_point (name: str) -> None

        : Method for set current point for generation code
    
    - pipeline (pipeline: list[str]) -> None

        : Set pipeline for get code
    
    - get_code () -> list[str]

        : Get full code builded by pipeline
    
    - get_str () -> str

        : Get full str code builded by pipeline

    - _append (code: list[str]) -> None
        
        : Method for append code for current point

    - _log(text: str, status: str) -> None

        : Method for add log
    '''
    def __init__(self) -> None:
        super().__init__('generatable')
        self.POINTS: dict[str, dict[str, list[str]]] = {
            'main': {'code': [], 'logs': []}
        }
        self.BUILDCONF: dict[str, list | dict] = {'pipeline': []}
        self.TEMP: dict[str, Any] = {'POINT': 'main', 'LOGGER': Logger()}
    
    def _log(self, text: str, status: str = 'MESSAGE'):
        log = self.TEMP['LOGGER'].format(text, status)
        self.POINTS[self.TEMP['POINT']]['logs'].append(log)
    
    def new_point(self, name: str) -> None:
        self.POINTS[name] = {'code': [], 'logs': []}
        self._log(f'created point {name}')
    
    def set_point(self, name: str) -> None:
        if name not in self.POINTS:
            raise KeyError(f"Point '{name}' doesn't exist")
        self.TEMP['POINT'] = name
        self._log(f'switched to point {name}')
    
    def _append(self, code: list[str]) -> None:
        self.POINTS[self.TEMP['POINT']]['code'].extend(code)
    
    def pipeline(self, pipeline: list[str]) -> None:
        self.BUILDCONF['pipeline'] = pipeline
        self._log(f'created pipeline {pipeline}')
    
    def get_code(self) -> list[str]:
        if not self.BUILDCONF['pipeline']:
            return ""
        
        ppl = self.BUILDCONF['pipeline']
        code_lines = []

        for point in ppl:
            if point not in self.POINTS:
                raise PipelineError(f'Unknown pipeline node: {point} (all: {self.POINTS})')
            code_lines.extend(self.POINTS[point]['code'])
        
        self._log(f'get code')
        
        return code_lines
    
    def get_str(self) -> str:
        return '\n'.join(self.get_code())









class builder(instruction):
    '''
    # Framework build metaclass-instruction
    Metclass for commands and instructions for Verge.

    Methods:
    - instruction (ins: object) -> None : add instruction
    '''

    def __init__(self) -> None:
        super().__init__('builder')
        self.INSTRUCTIONS: list[Any] = []
        self.name = 'generatable'

    def instruction(self, ins: object) -> None: 
        self.INSTRUCTIONS.append(ins)







def UVObj(obj_type: str = 'generatable'):
    '''
    # Basic metaclasses of framework

    Get basic unitverge metaclass.

    Types:
    - generatable: basic context metaclass
    - builder: basic instructions orchestrator
    - instruction: basic metaclass for Verge instructions
    '''
    if obj_type == 'generatable':
        return generatable
    elif obj_type == 'builder':
        return builder
    elif obj_type == 'instruction':
        return instruction
    else:
        raise ValueError(f'Unknown UVObj type: {obj_type}.')


from untvgdev.core.bx1.UnitBytexCore import (
    OutputError, 
    MemorySystemError, 
    PointError, 
    UnknownSyntax,
    VirtualMachineBasics
)


from untvgdev.core.bx1.UnitBytexDevConfig import (
    Debug
)



debugmode = Debug.executor
def print_debug(*data):
    if debugmode: print('debug:', *data)










class VirtualMachine:
    def __init__(self):
        self.vm = VirtualMachineBasics()
    
    def compile_type(self, data):
        data = str(data).replace('~', ' ').replace('``', '~')
        if data.isdigit():
            return int(data)
        try:
            return float(data)
        except ValueError:
            return data

    def compile_int(self, integer):
        inp = float(str(integer)
                  .replace(
                      '?', str(self.vm.position)
                      )
                  .replace(
                      '<', str(self.vm.tape[max(0, self.vm.position-1)])
                      )
                  .replace(
                      '>', str(self.vm.tape[self.vm.position+1])
                    )
                  .replace(
                      '!', str(self.vm.get())
                      ))
        return inp
    
    def jump(self, pos):
        pos = int(pos)
        self.vm.jump(pos)
    
    def set(self, data):
        data = self.compile_type(data)
        self.vm.set(data)
    
    def next(self, iters: int = 1):
        iters = int(iters)
        for i in range(iters):
            self.vm.next()
    
    def prev(self, iters: int = 1):
        iters = int(iters)
        for i in range(iters):
            self.vm.prev()

    def out(self): # out 
        print(self.vm.get(), end='')
    
    def otn(self): # out next
        print('\n', end = '')
    
    def otc(self): # out char
        try: print(
            chr(int(self.vm.get())), 
            end=''
        )
        except TypeError:
            OutputError(f'Invalid data of section to out char: {self.vm.get()}.')
    
    def ott(self): # out type now
        print(type(self.vm.get()), end='')
        
    def otp(self): # out current position
        print(self.vm.position, end='')
    
    def print(self, data):
        print(
            self.compile_type(data), 
            end=''
        )
    
    def instr(self): # input string
        data = input()
        try: 
            data = str(data)
            self.vm.set(data)
        except TypeError:
            OutputError(f'Invalid for input: {data} - type must be str.')
    
    def inint(self): # input integer
        data = input()
        try: 
            data = int(data)
            self.vm.set(data)
        except TypeError:
            OutputError(f'Invalid for input: {data} - type must be int.')
        
    def innum(self): # input any numeric data
        data = input()
        try: 
            data = float(data)
            self.vm.set(data)
        except TypeError:
            OutputError(f'Invalid for input: {data} - type must be number.')
    
    def plus(self, x): # plus any data
        self.vm.tape[
                self.vm.position
        ] += self.compile_type(x)
    
    def plus_int(self, x):
        self.vm.tape[
                self.vm.position
        ] += self.compile_int(x)

    def sub(self, x):
        self.vm.tape[
                self.vm.position
        ] -= self.compile_int(x)
    
    def div(self, x):
        self.vm.tape[
                self.vm.position
        ] /= self.compile_int(x)
    
    def mul(self, x):
        self.vm.tape[
                self.vm.position
        ] *= self.compile_int(x)
    
    def pow(self, x):
        self.vm.tape[
                self.vm.position
        ] **= self.compile_int(x)

     


























class Machine(VirtualMachine):
    def __init__(self):
        super().__init__()
        self.variables = {}   
    
    def new_var(self, name, value, type: str = 'basic'):
        value = self.compile_data(value)
        self.variables[name] = {'val': value, 'typ': type}
        
    def get_from_var(self, name):
        if name in self.variables:
            return self.variables[name]
        else:
            raise MemorySystemError(
        f"Variable '{name}' is not found."
            )

    def set_var(self, name, value):
        value = self.compile_data(value)
        if name in self.variables:
            var = self.variables[name]
            if var['typ'] != 'const':
                self.variables[name]['val'] = value
            else:
                raise MemorySystemError(
        f"Variable '{name}' is not muttable."
            )
        else:
            raise MemorySystemError(
        f"Variable '{name}' is not found."
            )

    def compile_data(self, data):
        data = str(data)
        if data.startswith('$'):
            if data[1:] in self.variables:
                return self.variables[data[1:]]['val']
        elif data.startswith('%'):
            data = self.compile_path(data[1:])
            if isinstance(data, int) and \
            data in range(len(self.vm.tape)):
                return self.vm.tape[data]
        elif data == '!':
            return self.vm.get()
        elif data == '?':
            return self.vm.position
        elif data == '<':
            return self.vm.tape[self.vm.position-1]
        elif data == '>':
            return self.vm.tape[self.vm.position+1]
        return self.compile_type(data)

    def compile_path(self, data):
        data = str(data)
        if data.startswith('$'):
            if data[1:] in self.variables:
                return self.variables[data[1:]]['val']
        elif data == '!':
            return self.vm.get()
        elif data == '?':
            return self.vm.position
        elif data == '<':
            return self.vm.position-1
        elif data == '>':
            return self.vm.position+1
        return self.compile_type(data)
    
    def swap(self, pos):
        pos = self.compile_data(pos)
        curr = self.vm.get()
        oth = self.vm.tape[pos]
        self.vm.set(oth)
        self.vm.tape[pos] = curr
    
    def copy(self, pos):
        pos = self.compile_data(pos)
        curr = self.vm.get()
        self.vm.tape[pos] = curr





























from .backend.back import Generator

class Frontend:
    def __init__(self) -> None:
        self.gen = Generator()
    
    def query(self, line: str):
        self.gen.py(line)
        return self.gen.build()

    def load(self, name: str):
        self.query(f'MEM2.load({repr(name)})')
        return self
    
    def save(self, name: str):
        self.query(f'MEM2.save({repr(name)})')
        return self
    
    def clear(self):
        self.query(f"MEM2.clear()")
        return self
    
    def setto(self, key, data, repricate = True):
        self.query(
            f"MEM2.setto(key = {repr(key)}, data = {repr(data) if repricate else data})"
        )
        return self
    
    def getfrom(self, key, out = 'res'):
        self.query(f"{out} = MEM2.getfrom({repr(key)})")
        return self
    
    def select(self, condition: str, ):
        code = f'''
result = {{}}
try:
    for key in MEM2.keys():
        value = MEM2.getfrom(key)
        if {condition.replace('$', 'value')}:
            result[key] = value
except IndexError as e:
    print(f'\\nSelect Error:\\n{{e}}')
        '''
        self.query(code)
        return self
    
    def exec(self):
        lclscp = locals() # localscope
        exec('\n'.join(self.gen.build()), lclscp, lclscp)
        return lclscp





class Query(Frontend):
    def __init__(self, database: str) -> None:
        super().__init__()
        self.db = database
        self.scope = {}
    def __enter__(self):
        self.load(self.db)
        return self
    def __exit__(self, *args, **kwars):
        self.save(self.db)
        self.scope = self.exec()


def NewDB(name: str):
    Frontend().save(name).exec()


from .Frontend import Frontend, Query


class UnQL:
    def __init__(self, database: str = 'unitdb', format: str = 'undb') -> None:
        self._f = Frontend()
        self.name = f"{database}.{format}"
    
    def request(self, type = 'GET', loc = locals(), **kwargs):
        type = type.lower()
        loc['Query'] = Query
        loc['Frontend'] = Frontend
        def check(length, required: list[str] = ['Not Stated']):
            if len(kwargs.keys()) != length:
                raise ValueError(
                    f'Invalid argument length. Required: {", ".join(required)} ({len(required)}). Got: {len(kwargs.keys())}'
                )
        match type:
            case 'get':
                check(1, ['FROM'])
                RESULT = '_res'
                code = f'''
with Query({repr(self.name)}) as q:
    q.getfrom({repr(kwargs['FROM'])}, out={repr(RESULT)})
{RESULT} = q.scope[{repr(RESULT)}]
                '''
                exec(code, loc, loc)
                return loc[RESULT]
            case 'put':
                check(2, ['TO', 'DATA'])
                code = f'''
with Query({repr(self.name)}) as q:
    q.setto(key = {repr(kwargs['TO'])}, data = {repr(kwargs['DATA'])})
                '''
                exec(code, loc, loc)
                return self
            case 'select':
                check(1, ['WHEN'])
                RESULT = '_result'
                code = f'''
with Query({repr(self.name)}) as q:
    q.select(condition = {repr(kwargs['WHEN'])})
{RESULT} = q.scope[{repr(RESULT)}]
                '''
                exec(code, loc, loc)
                return loc[RESULT]
            case _:
                raise SyntaxError(f'Unknow request type: {type}')
        
    def __call__(self, *args, **kwargs):
        return self.request(*args, **kwargs)

    def __getitem__(self, __i):
        return self.request('GET', FROM=__i)
    
    def __setitem__(self, __i, __o):
        self.request('PUT', TO=__i, DATA=__o)


'''
# ByteX

UnitBytex language.

System assembler-like language for UnitVerge framework programms.

Executing in Vurtual machine.

By pt.

'''

__version__ = '2.0'


def get_from_ver(__version__: str = '2.0'):
    if __version__ == '1.0':
        from .bx1 import lang
    elif __version__ == '2.0':
        from .bx2 import lang
    else: 
        print(f'ByteX {__version__} not allwoed now. Exiting...')
        exit()
    return lang

lang = get_from_ver('2.0')


import time


class Logger:
    '''
    # Simple logger
    Simple logger for code generation in Unit.

    Methods:
    - format (text: str, status: str) -> str

        : Format text to log
    
    - log (text: str, status: str) -> None
    
        : Orint formatted log
    '''
    def __init__(self):
        self.logs = []

    def format(self, text: str = 'empty', status: str = 'MESSAGE') -> str:
        return (f'LOG: [{time.time()}] : [{status}] : [{text}]')
    
    def log(self, text: str, status: str = 'MESSAGE') -> None:
        text = self.format(text, status)
        self.logs.append(text)
        print(text)


from .Basics._base import UVObj




class Verge:
    '''
    # Verge
    KeyClass for interprete instructions.

    ## Methods:
    - `interprete(instructions)` : interprete list of instructions.
    - `compile` : get verge code.
    - `__call__(instructions)` : interprete list of instructions and get result.
    - `other systems...`

    ## Example:
    >>> v = Verge()
    >>> v.interprete(editor, unit1, unit2)
    >>> final_code = v.compile()
    >>> local = Bytex2().current['ex'](final_code)

    ### or just
    >>> final_code = Verge()(editor, unit1, unit2)
    >>> local = Bytex2().current['ex'](final_code)
    '''
    def __init__(self, **kwargs) -> None:
        self.instruction = {
            'generatable':{'type':'build', 'getcode':'get_code'},
            'builder':{'type':'edit', 'inst_attr':'INSTRUCTIONS'},
        }
        self.type = {
            'build': self.build,
            'edit': self.edit,
        }
        self.code = [] \
            if '__code' not in kwargs.keys() else kwargs['__code'] # for repr
    
    def interprete(self, instructions: list) -> Verge:
        '''
        # Interprete
        interprete list of instructions method
        '''
        if not isinstance(instructions, (list, tuple)):
            raise TypeError(f"instructions must be list, got {type(instructions)}")
        ins: object # just type hint for editor
        for ins in instructions:
            # all instructions must be extends from [instruction] metaclass
            # instruction metaclass has [__repr__] and other methods to confirm this checks
            ins_type = repr(ins)
            if not ins_type.startswith('instruction-'):
                raise ValueError(f'Bad instruction {repr(ins)}')
            try: ins_type = ins.name
            except: raise ValueError(f'Bad instruction {repr(ins)} : has no attribute name.')
            if ins_type in self.instruction.keys():
                # interprete instrucrion
                self.work(ins, ins_type)
            else: 
                raise ValueError(f'Instruction {repr(ins)} ({ins_type}) not in Verge instructions')
        return self
    
    def work(self, instruction: object, typ: str) -> Verge:
        'interprete one instruction method'
        # get instruction type
        _type = self.instruction[typ]['type']
        try: # try get handler for instruction by type
            handler = self.type[_type] 
        except:
            raise ValueError(f'Bad instruction {repr(instruction)} : unknown type {_type} : has no handler.')
        handler(instruction, typ)
        return self

    def build(self, ins: object, typ: str) -> Verge:
        'handler for build-type instructions'
        try: # try to get name of method to get all code
            method = self.instruction[typ]['getcode'] 
        except: raise ValueError(f'Bad instruction {repr(ins)} : has no key getcode with name of attribute.')
        # get attribute for get code
        attr = ins.__getattribute__(method)
        try: # try get code
            code = attr()
        except Exception as e:
            raise ValueError(f'Bad instruction {repr(ins)} : error in handler : {e}')
        if not isinstance(code, list):
            raise ValueError(f'Bad handler : returning type is {type(code)}, but must be list')
        # append code to self.code
        for line in code: self.code.append(line)
        return self
    
    def edit(self, ins: object, typ: str) -> Verge:
        'handler for type '
        try:

            method = self.instruction[typ]['inst_attr']
        except: raise ValueError(f'Bad instruction {repr(ins)} : has no key INSTRUCTIONS with list of instructions.')
        attr = ins.__getattribute__(method)
        if not isinstance(attr, list):
            raise ValueError(f'Bad instruction {repr(ins)} : INSTRUCTIONS must be list, not {type(attr)}.')
        for sub_ins in attr:
            try: sub_ins(self)
            except Exception as e: 
                raise ValueError(f'Bad instruction {repr(sub_ins)} : bad INSTRUCTIONS for Verge : {e}.')
        return self

    def compile(self) -> list[str]:
        '''
        # Compile
        ### -> CODE ( list[str] )
        Get code of Verge. Returns list of str with interpreted code.
        '''
        return self.code
    
    def __call__(self, instructions: list[str]) -> list[str]:
        '''
        # Interprete and get code
        ### -> CODE ( list[str] )
        'UI' for verge. Interprete list of instructions and get code.
        '''
        self.interprete(instructions)
        return self.compile()
    
    def __repr__(self):
        return f'__code = {repr(self.code)}; Verge(__code = __code)'



class SForm:
    '''
    # Syntax Formater
    Simple formater values and others for bytex.
    '''
    def text(value: str) -> str:
        '''
        ### Simple text format

        Only for simplify code read (returning repr value)
        '''
        return str(repr(value))
    
    def value(value):
        '''
        ### Simple value format

        Only for simplify code read (returning raw value)
        '''
        return value


from ._base import generatable


class _UnitMeta(generatable):
    '''
    Basic Unit metaclass and functions.
    '''

    def __init__(self) -> None:
        super().__init__()
        self._counter = 0

    @property
    def _temp(self) -> int:
        'system'
        # method to get counter
        self._counter += 1
        return self._counter

    @property
    def _unique_name(self):
        'system'
        # method to get auto-id
        return f"__auto_{self._counter}_{id(self)}"

    def condblock(Self, cond_type: str = 'if'):
        '''
        Create condition block.

        Condition types:

        - if
        - elif
        - else
        '''
        if cond_type not in ('if', 'elif', 'else'):
            raise ValueError(f"Invalid condition type [{cond_type}] (can be only if/elif/else)")

        class _cond:
            def __init__(self, condtition: str) -> None:
                '''
                # Condition
                Create condition

                ## Args:
                - condition: str
                '''
                self.localself = Self
                self.cond = condtition
                self.__name = self.localself._temp

            def __enter__(self):
                self.localself.raw(f'stdfunc cond{self.__name} MEM, LANG')
                return self

            def __call__(self, *args: any, **kwds: any) -> 'Unit':
                return self.localself

            def __exit__(self, *args, **kwargs):
                self.localself.raw([
                    f'end',
                    f'var __cond{self.__name} {self.cond}',
                    f'src {cond_type} __cond{self.__name}:',
                    f'$addtabs',
                    f'do cond{self.__name} MEM, LANG',
                    f'$subtabs',
                ])

        return _cond

    @property
    def function(Self):
        '''
        Create function block.
        '''

        class _func:
            def __init__(self, name: str, args: list[str] = [], type: str = 'std') -> None:
                '''
                # Function
                Create function.

                ## Args:
                - name: str - name of function
                - args: list[str] - list of arguments.

                    format can be: ['arg', 'arg: int', 'arg: int = 10']

                - type: str - type of function. Types:

                    - 'std' : default function

                    - 's' : stctic class method

                    - 'cls' : class function (1st argument - self)
                '''
                self.name = name
                self.type = type
                self.args = args
                self.localself = Self

            def __enter__(self):
                self.localself.raw(
                    f'{self.type if self.type != "cls" else ""
                    }func {self.name} {", ".join(self.args)}'
                )
                return self

            def __exit__(self, *args, **kwargs):
                self.localself.raw(f'end')

            def call(self, *args: str, variable_name='result') -> str:
                '''
                Call function and create variable in code with result of function.

                -> str : name of variable
                '''
                self.localself.raw(f'src {variable_name} = {self.name}({", ".join(args)})')
                return variable_name

            def __call__(self, *args: any, **kwds: any) -> 'Unit':
                return self.localself

        return _func

    @property
    def classblock(Self):
        '''
        Create class block.
        '''

        class _class:
            def __init__(self, name: str, extends: str | None = None) -> None:
                '''
                # Class
                Create class.

                ## Args
                - name: str - name of class
                - extends: str - name of metaclass
                '''
                self.meta = extends
                self.name = name
                self.localself = Self

            def __enter__(self):
                self.localself.raw(f'struct {self.name}({self.meta})')
                return self

            def __call__(self, *args: any, **kwds: any) -> 'Unit':
                return self.localself

            def __exit__(self, *args, **kwargs):
                self.localself.raw('endstruct')

        return _class

    @property
    def iterable(Self):
        '''
        Create iterable block.
        '''

        class _iters:
            '''
            # Iterable
            Create block with repeats.

            ## Args:
            - iters: int - repeats count
            '''

            def __init__(self, iters: int = 0) -> None:
                self.iters = int(iters)
                self._id = Self._temp
                self.localself = Self

            def __enter__(self):
                self.localself.raw(f'stdfunc iters{self._id}')
                return self

            def __call__(self, *args: any, **kwds: any) -> 'Unit':
                return self.localself

            def __exit__(self, *args, **kwargs):
                self.localself.raw(f'end')
                self.localself.raw(f'iters iters{self._id} {self.iters}')

        return _iters

    def raw(self, code: list[str] | str):
        '''
        Add raw bytex2-code to context.
        '''
        if isinstance(code, list):
            self._append(code)
        elif isinstance(code, str):
            self._append([code])
        else:
            raise ValueError('Unknown input type')
        return self

    def python(self, code: str):
        '''
        Add raw python to context.
        '''
        if isinstance(code, str):
            code = code.replace('    ', '\\tab|')
        else:
            raise ValueError('Unknown input type')
        self.raw(['python'] + code.split('\n') + ['python'])
        return self

    def start_context(self, name: str):
        'system'
        self.raw(f"stdfunc {name}")
        return self

    def end_context(self):
        'system'
        self.raw('end')
        return self


class _UnitFrontendMeta(_UnitMeta):
    '''
    Unit aliases and simples metaclass.
    '''

    @property
    def ifblock(self):
        'Get if block'
        return self.condblock('if')

    @property
    def elifblock(self):
        'Get elif block'
        return self.condblock('elif')

    @property
    def elseblock(self):
        'Get else block'
        return self.condblock('else')

    def delay_btx(self, time_ms: int = 0):
        'Delay method'
        if not isinstance(time_ms, int):
            raise ValueError(f"Bad input for delay: [{type(time_ms)}{time_ms}], muste be int.")
        self.raw(f'delay {time_ms}')
        return self

    def spawn_btx(self, name: str):
        'Start thread method'
        self.raw(f'spawn {name}')

    def import_py(self, module: str):
        'Import pyhton lib to bytex2 method'
        if not isinstance(module, str):
            raise ValueError(f"Bad input for delay: [{type(module)}{module}], muste be int.")
        self.raw(f'#import {module}')
        return self

    def debug_btx(self):
        'Simple debuger (work only in runtime)'
        self.raw('debug')
        return self

    def callfunc_btx(self, func_name: str, *args: str, return_var: str = 'result'):
        '''
        Call function and create variable in code with result of function.

        -> str : name of variable
        '''
        self.raw(f'src {return_var} = {func_name}({", ".join(args)})')
        return return_var

    def inputint_btx(self, prompt: str = ''):
        'Input integer method'
        self.println(prompt)
        self.raw('in')
        return self

    def var_btx(self, name: str, value):
        'Create variable method'
        self.raw(f"var {name} {value}")
        return self

    def println_btx(self, text, to_repr: bool = False):
        'Println method'
        text = repr(text) if to_repr else text
        self.raw(f"echoln {text}")
        return self

    def print_btx(self, text, to_repr: bool = False):
        'Print method'
        text = repr(text) if to_repr else text
        self.raw(f"echo {text}")
        return self

    def loadto_sec(self, value: int):
        'Load integer to current section method'
        if not isinstance(value, int):
            raise ValueError(f"Bad input to load: [{type(value)}{value}], muste be int.")
        self.raw(f"load {int(value)}")
        return self

    def set_hand(self, handname: str):
        'Set current hand method'
        self.raw(f"switch {handname}")
        return self

    def new_hand(self, handname: str):
        'Create new hand method'
        self.raw(f"new {handname}")
        return self

    def next_reg(self):
        'Jump to next register method'
        self.raw('>')
        return self

    def prev_reg(self):
        'Jump to prev register method'
        self.raw('<')
        return self

    def jumpto_reg(self, regindex: int):
        'Jump to registre method'
        if not isinstance(regindex, int):
            raise ValueError(f"Bad input to jump: [{type(regindex)}{regindex}], muste be int.")
        self.raw(f"moveto {int(regindex)}")
        return self

    def jumpto_sec(self, secindex: int):
        'Jump to section in register method'
        if not isinstance(secindex, int):
            raise ValueError(f"Bad input to jump: [{type(secindex)}{secindex}], muste be int.")
        self.raw(f"jump {int(secindex)}")
        return self

    def operation_sec(self, value: int, operation: str, copyto_regsec: str | None = '0:0'):
        'Do operation and copy value to register'
        # check types
        if not isinstance(value, int):
            raise ValueError(f"Bad input to opertion: [{type(value)}{value}], muste be int.")
        if not isinstance(operation, str):
            raise ValueError(f"Bad input to opertion: [{type(operation)}{operation}], muste be int.")
        if copyto_regsec is not None and not isinstance(copyto_regsec, str):
            raise ValueError(f"Bad input to opertion: [{type(copyto_regsec)}{copyto_regsec}], muste be int.")
        operations = {
            'add': 'add', 'plus': 'add', '+': 'add',
            'sub': 'sub', 'subdivission': 'sub', 'minus': 'sub', '-': 'sub',
            'div': 'div', 'divission': 'div', '/': 'div',
            'mul': 'mul', 'mull': 'mul', 'multyply': 'mul', '*': 'mul',
            'pow': 'pow', 'power': 'pow', '**': 'pow', '^': 'pow',
        }
        raw_op = operation
        operation = operations.get(operation)
        # check operation
        if operation is None:
            raise ValueError(f'Unknown operation: [{raw_op}].')
        if operation == 'add' and value == 1:
            self.raw(['+'])
        if operation == 'sub' and value == 1:
            self.raw(['-'])
        self.raw([
            f"var __currh MEM.hand",
            f"new TEMP",
            f"switch TEMP",
            f"reg -1",
            f"var __currr LANG.register",
            f"moveto -1",
            f"jump 4095",  # jump to last section
            f"load {int(value)}",
            f"jump 0",
            f"{operation} MEM.POINTER 4095",
            f"copy {copyto_regsec}" if copyto_regsec else ''
        ])
        return self

    def save_context(self, name: str):
        'Save current context'
        self.raw(f'save "{name}.btxf"')
        return self

    def open_context(self, name: str):
        'Load to current context'
        self.raw(f"open {name}.btxf")
        return self

    def append_context(self, name: str):
        'Append bytex file to bytex code'
        self.raw(f'#append {name}')
        return self


class Unit(_UnitFrontendMeta):
    '''
    # Unit
    Code context class.

    Contain all code-generation methods.
    '''

    def __init__(self, name: str = 'unk') -> None:
        super().__init__()
        self.plugins = {}
        self.contextname = name

    def __enter__(self):
        self.start_context(self.contextname)
        return self

    def __exit__(self, *args, **kwargs):
        self.end_context()

    def plugin(self, name: str, handler: object):
        '''
        Create your plugin (Method) in unit.
        Handler first argument is [Self] object of Unit, other is user args.
        '''
        if hasattr(self, name):
            raise ValueError(f"Plugin name '{name}' conflicts with built-in method")
        self.plugins[name] = handler
        return self

    def register_plugin(self, name: str):
        'Create your plugin by function and name.'
        def handler(func):
            self.plugin(name, func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return handler

    def compile(self) -> str:
        'Get str-code of unit'
        return self.get_str()

    def __getattr__(self, name):
        try:
            rawattr = self.plugins[name]

            def attr_plugin_handler(*args, **kwargs):
                return rawattr(self, *args, **kwargs)

            return attr_plugin_handler
        except:
            try:
                return self.__getattribute__(name)
            except AttributeError as e:
                raise AttributeError(f"Unknown attrubute '{name}'. \n(Error: {e})")

    def __call__(self, line: str):
        return self.raw(line)


class RuntimeError(Exception):
    def __init__(self, error, text = '', line = [], code = []):
        super().__init__(
            (f"\nBYTEX2 Runtime Error:" ) +
            (f"\n{error}" if error else '') +
            (f"\nerr: {text}") +
            (f"\n     line: {" ".join(line)}" if line else '') +
            (f"\n     line: {code.index(line) + 1}" if line else '')
        )
        
class TranslationError(RuntimeError):
    def __init__(self, error = '', text='', line=[], code=[], type: int = 0):
        super().__init__(f'TranslationError: {type}' if error == '' else error, 
                         text, line, code)

class MemoryError(RuntimeError):
    def __init__(self, error = '', text='', line=[], code=[], type: int = 0):
        super().__init__(f'MemoryError: {type}' if error == '' else error, 
                         text, line, code)

class Bytex2Error(RuntimeError):
    def __init__(self, error = '', text='', line=[], code=[], type: int = 0):
        super().__init__(f'Bytex2Error: {type}' if error == '' else error, 
                         text, line, code)





from .translator.translator import Translator
from .translator import __version__





def build(code: str):
    translated = '\n'.join(Translator().translate(code))
    return translated




def bin(code: str, fileout: str = 'a.out'):
    built = build(code)
    with open(fileout, 'w') as f:
        f.write(built)
    return built
    

def execute(code: str):
    localscope = locals()
    exec(build(code), localscope, localscope)
    return localscope


'''
# UnitVerge Framework
'''

__version__ = '0.9.5'

'UV - UNITVERGE'
from .Verge import Verge as Verge
from .Verge.Basics.syntax import SForm as SForm
from .Verge import UVObj
from .Verge import Unit as Unit





class Bytex:
    '''
    # Bytex lang content.
    ## Methods:
    - `get()` -> `dict[str, object]` : get `_current` (from parameter ver) version lang 
    - `current` -> `dict[str, type]` : get bytex2 lang version
    - `exec_main(code, main_method)` : `dict[object, Any]` : execute code by method
    - `translator()` -> `type[Translator]` : get language translator (bytex2)
    '''
    _current = '2.0'
    def get(self):
        'get lang with _current parameter version'
        from .Bytex import get_from_ver
        return get_from_ver(self._current)()
    @property
    def current(self) -> dict[str, type[any]]:
        'get bytex2 lang version'
        from .Bytex.bx2 import lang
        return lang()
    def exec_main(
        self, 
        code: list [str] | str, 
        main_method: str = 'main'
    ) -> dict[str, any]:
        'execute bytex code with main method'
        if isinstance(code, list): code = '\n'.join(code)
        code += f'\ndo {main_method}'
        return self.current['ex'](code)
    @property
    def translator(self):
        'get bytex2 translator'
        from .Bytex.bx2 import Translator
        return Translator




def compile(VergeObject: Verge.Verge) -> dict[str, object]:
    '''
    # Compile and run code
    Compile and run code (bytex2) from Verge and get locals.
    '''
    # get code from verge
    code = '\n'.join(VergeObject.compile())
    # execute btx2 code and get localscope
    locals = Bytex().current['ex'](code)
    # return localscope
    return locals



from . import __version__
import time

core = f'''

class Memory:...



import threading
import pickle
import time
import abc
from UnitVerge.Bytex.bx2.BYTEX2_back import Memory, Memory2


class VMEXCEPTION(Exception): pass
class EXECUTIONERROR(Exception): pass



try: from numba import jit
except:
    def jit(*args, **kwargs):
        def wrap(func):
            return func
        return wrap
    



# METADATA
meta = """
date: {time.time()}
version = {__version__}
development by pt, 2025
"""


points = []
libs = []
ver = {__version__}
_is_lib = False
MEM = Memory()
MEM2 = Memory2()



class language:
    _ = 0
    @property
    def register(self):
        return MEM.CURSOR
    @property
    def section(self):
        return MEM.POINTER
    @property
    def getdata(self):
        return MEM.getdata()
    @property
    def nextreg(self):
        return MEM.CURSOR + 1
    @property
    def prevreg(self):
        return MEM.CURSOR - 1
    @property
    def nextsec(self):
        return MEM.POINTER + 1
    @property
    def prevsec(self):
        return MEM.POINTER - 1
    def haslib(self, name: str):
        return name in libs
    def getp(self, index: int):
        return MEM.getfrom_p(int(index))
    def setp(self, index: int, data):
        return MEM.setto_p(int(index), data)

LANG = language()

'''.split('\n')


from setuptools import setup, Extension, find_packages

__bx_version__ = '2.0'
__framework_verion__ = '1.1.3'

match __bx_version__:
    case '1.0':
        path = 'src/C-CORES/bx1/'
        bytex = Extension(
            name='untvgdev.core.bx1.UnitBytexCore',
            sources=[path + 'UnitBytexCore.c'],
        )
        config = Extension(
            name='untvgdev.core.bx1.UnitBytexDevConfig', 
            sources=[path + 'UnitBytexDevConfig.c'],
        )
        c_mod = [bytex, config]
    
    case '2.0':
        path = 'src/UnitVerge/Bytex/bx2/'
        basemem = Extension(
            name='BYTEX2_back',
            sources=[path + 'machine/BYTEX2_back.c'],
        )
        c_mod = [basemem]


setup(
    name='unitverge',
    author='Pt',
    author_email='kvantorium73.int@gmail.com',
    version=__framework_verion__,
    ext_modules=c_mod,
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    python_requires='>=3.10',
    install_requires=[],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ],
    description='A metaprogramming framework for code generation and DSL creation',
    url='https://github.com/pt-main/unitverge',
    include_package_data=True,
    options={
        'bdist_wheel': {
            'python_tag': 'cp314',
            'plat_name': 'manylinux2014_x86_64',
        }
    }
)



'''
# UnitBytex language

System assembler-like language for UnitVerge framework programms.

Executing in Vurtual machine.

By pt.

## Bytex 1.0
'''



__version__ = '1.0'

from ._lang import execute, Executor

def lang():
    return {
        'bu': lambda *args, **kwrgs: None, 
        'bi': bin,
        'ex': execute,
        'tr': lambda *args, **kwrgs: None
    }


from untvgdev.core.bx1.UnitBytexCore import Parser, InterpreteError
from .__executor import Executor
import sys

# for cycles
sys.setrecursionlimit(16384)



class Interpreter:
    def __init__(self):
        self.exec = Executor()
        self.points_code = {}
        self.points = []
    
    def execute(self, code):
        code_spl = Parser(code) # splitted code
        for line in code_spl:
            try: 
                result = self.exec.exec(line)
            except Exception as e:
                raise InterpreteError(f"\n{e.__class__}: {e}\nLine {code_spl.index(line)}:\n>  {' '.join(line)}")
            if result.to_return != None: 
                return result.to_return
        return self




def execute(code: str) -> int | None:
    return Interpreter().execute(code)




if __name__ == '__main__':
    execute('''
point input; p:input jump 1234;
point basic; p:basic jump 0;
point temp; p:temp jump 8421;
point main;
    p:main goto temp;
    p:main copy 1235;
    p:main goto input;
    p:main op * >;
    p:main out .;
goto temp; print Введите число 1:~; in num; goto basic;
goto input; print Введите число 2:~; in num; goto basic;
goto main;
>;
    ''')


from .__vm import *
import sys, gc


class Executor(Machine):
    def __init__(self):
        super().__init__()
        self.points_code = {}
        self.points = []
        self.add_to_point = [False, '']
        self.to_return = None



    def exec(self, command: list):
        cmd = command[0]
        args = command[1:]

        if self.add_to_point[0] and \
        not cmd.startswith('pe:'):
            args = [cmd] + args
            cmd = f'p:{self.add_to_point[1]}'
            print_debug(cmd, args)

        def err_syntax(other = ''):
                raise UnknownSyntax(f"Unknown syntax: {' '.join(command)}. " + str(other))
        def check_args_len(length):
            if len(args) != length:
                err_syntax(
                    f'\nBad arguments length ({len(args)} while must be {length}).'
            )

        print_debug(f'Input - {command}, cmd - {cmd}, args - {args}')

        if cmd.startswith('p:'):
            name = cmd[2:]
            if name in self.points:
                # add to point code from command
                self.points_code[name].append(args)
            else:
                # if name of point is not found
                raise PointError(f'Unknown name to add code to point: {name}')
        elif cmd.startswith('ps:'): # point block start
            name = cmd[3:]
            if name in self.points:
                self.add_to_point[0] = True
                self.add_to_point[1] = name
            else:
                raise PointError(f'Unknown name to start point block: {name}')
        elif cmd.startswith('pe:'): # point block end
            name = cmd[3:]
            if name in self.points:
                self.add_to_point[0] = False
                self.add_to_point[1] = ''
            else:
                raise PointError(f'Unknown name to end point block: {name}')
        else:
            match cmd:
                case 'jump':
                    check_args_len(1)
                    arg = self.compile_path(args[0])
                    self.jump(arg)

                case 'set':
                    check_args_len(1)
                    arg = self.compile_data(args[0])
                    self.set(arg)

                case 'return':
                    self.to_return = self.vm.tape[
                    self.vm.position
                ]

                case 'load':
                    if len(args) < 2:
                        err_syntax(
                        f'\nBad arguments length ({len(args)} while must be 2 and more).'
                    )
                    pos = self.compile_path(args[0])
                    data = self.compile_data(' '.join(args[1:]))
                    pre = self.vm.position
                    self.vm.position = pos
                    self.set(data)
                    self.vm.position = pre


                case '>':
                    self.next()

                case '<':
                    self.prev()

                case 'next':
                    check_args_len(1)
                    arg = self.compile_data(args[0])
                    self.next(arg)

                case 'prev':
                    check_args_len(1)
                    arg = self.compile_data(args[0])
                    self.prev(arg)

                case 'out':
                    check_args_len(1)
                    arg = args[0]
                    if arg in ('.', 'data', 'd'):
                        self.out()
                    elif arg in ('next', 'n'):
                        self.otn()
                    elif arg in ('char', 'c'):
                        self.otc()
                    elif arg in ('type', 't'):
                        self.ott()
                    elif arg in ('pos', 'p', '?'):
                        self.otp()
                    else:
                        err_syntax(f'Unknown out type: {arg}.')

                case 'op': # operation
                    check_args_len(2)
                    arg = args[0]
                    x = self.compile_data(args[1])
                    if arg in ('plus'):
                        self.plus(x)
                    elif arg in ('+'):
                        self.plus_int(x)
                    elif arg in ('-'):
                        self.sub(x)
                    elif arg in ('/'):
                        self.div(x)
                    elif arg in ('*'):
                        self.mul(x)
                    elif arg in ('^'):
                        self.pow(x)
                    else:
                        err_syntax(f'Unknown op type: {arg}.')

                case 'work':
                    self.workparser(args)

                case 'malloc':
                    check_args_len(1)
                    self.vm.recreate(
                        int(self.compile_data(args[0]))
                    )

                case 'rlim':
                    check_args_len(1)
                    sys.setrecursionlimit(
                        int(self.compile_data(args[0]))
                    )

                case 'in':
                    check_args_len(1)
                    arg = args[0]
                    if arg in ('str', 's'):
                        self.instr()
                    elif arg in ('int', 'i'):
                        self.inint()
                    elif arg in ('num', 'n'):
                        self.innum()
                    else:
                        err_syntax(f'Unknown input type: {arg}.')

                case 'print':
                    arg = self.compile_data(
                        ' '.join(args)
                    )
                    self.print(arg)

                case 'point': # declarate point
                    name = args[0]
                    self.points.append(name)
                    self.points_code[name] = [['#', 'start']]

                case 'goto': # use point
                    name = self.compile_data(args[0])
                    if name in self.points_code.keys():
                        for linelocal in self.points_code[name]:
                            self.exec(linelocal)
                    else:
                        raise PointError(f'Unknown name for go to point: {name}')

                case 'copy':
                    check_args_len(1)
                    arg = self.compile_path(args[0])
                    self.copy(arg)

                case 'swap':
                    check_args_len(1)
                    arg = self.compile_path(args[0])
                    self.swap(arg)

                case 'mem':
                    arg = args[0]
                    if arg == 'var':
                        check_args_len(3)
                        self.new_var(args[1], args[2])
                    elif arg == 'const':
                        check_args_len(3)
                        self.new_var(args[1], args[2], 'const')
                    elif arg == 'edit':
                        check_args_len(3)
                        self.set_var(args[1], args[2])
                    elif arg == 'temp':
                        self.temp_use(args[1], args[2:])
                    else:
                        err_syntax(f'Unknown mem arg: {arg}.')

                case '#':
                    pass

                case 'dlall': # delete all
                    gc.collect(0)
                    gc.collect(1)
                    gc.collect(2)

                case 'del': # delete trash
                    arg = args[0]
                    match arg:
                        case 'collect':
                            typ = args[1]
                            if typ == 'all':
                                gc.collect(0)
                                gc.collect(1)
                                gc.collect(2)
                            elif typ == 'trash':
                                gc.collect(2)
                            else:
                                gc.collect(
                                    int(self.compile_int(typ))
                                )
                        case 'off':
                            gc.disable()
                        case 'on':
                            gc.enable()
                        case _: err_syntax(f'Unknown del arg: {arg}.')

                case _:
                    if str(cmd).strip() != '':
                        err_syntax()
        return self



    def work_ife(self, condition, command: list): # work if equal
        condition = self.compile_data(condition)
        if self.vm.get() == condition:
            self.exec(command)

    def work_ifin(self, condition, command: list): # work if in
        condition = self.compile_data(condition)
        if self.vm.get() in condition:
            self.exec(command)

    def work_ifins(self, condition, command: list): # work if in (string version)
        condition = self.compile_data(condition)
        if str(self.vm.get()) in str(condition):
            self.exec(command)

    def work_ifne(self, condition, command: list): # work if not equal
        condition = self.compile_data(condition)
        if self.vm.get() == condition:
            return
        self.exec(command)

    def work_ifm(self, condition, command: list): # work if more
        condition = self.compile_data(condition)
        if self.vm.get() > condition:
            self.exec(command)

    def work_ifl(self, condition, command: list): # work if less
        condition = self.compile_data(condition)
        if self.vm.get() < condition:
            self.exec(command)

    def work_ifme(self, condition, command: list): # work if more or equal
        condition = self.compile_data(condition)
        if self.vm.get() >= condition:
            self.exec(command)

    def work_ifle(self, condition, command: list): # work if less or equal
        condition = self.compile_data(condition)
        if self.vm.get() <= condition:
            self.exec(command)

    def temp_use(self, var, command: list):
        var_name = self.compile_data(var)
        pos = self.vm.position
        self.jump(0)
        self.set(self.get_from_var(var_name)['val'])
        self.exec(command)
        self.jump(pos)

    def workparser(self, args):
        arg = args[0]

        if arg == 'ife':
            self.work_ife(args[1], args[2:])
        elif arg == 'ifin':
            self.work_ifin(args[1], args[2:])
        elif arg == 'ifins':
            self.work_ifins(args[1], args[2:])
        elif arg == 'ifne':
            self.work_ifne(args[1], args[2:])
        elif arg == 'ifl':
            self.work_ifl(args[1], args[2:])
        elif arg == 'ifm':
            self.work_ifm(args[1], args[2:])
        elif arg == 'ifme':
            self.work_ifme(args[1], args[2:])
        elif arg == 'ifle':
            self.work_ifle(args[1], args[2:])

        elif arg == 'whlne': # while not equal
            args[1] = self.compile_data(args[1])
            while self.vm.get() != args[1]:
                self.exec(args[2:])
        elif arg == 'whle': # while equal
            args[1] = self.compile_data(args[1])
            while self.vm.get() == args[1]:
                self.exec(args[2:])
        elif arg == 'whlm': # while more
            args[1] = self.compile_data(args[1])
            while self.vm.get() > args[1]:
                self.exec(args[2:])
        elif arg == 'whlnme': # while more or equal
            args[1] = self.compile_data(args[1])
            while self.vm.get() >= args[1]:
                self.exec(args[2:])
        elif arg == 'whll': # while less
            args[1] = self.compile_data(args[1])
            while self.vm.get() < args[1]:
                self.exec(args[2:])
        elif arg == 'whlle': # while less or equal
            args[1] = self.compile_data(args[1])
            while self.vm.get() <= args[1]:
                self.exec(args[2:])

        elif arg == 'repeat':
            iterations = int(
                self.compile_data(args[1])
            )
            for _ in range(iterations):
                self.exec(args[2:])
        else:
            raise UnknownSyntax(f'Unknown work type: {arg}.')


from UnitVerge.Bytex import lang
import time
execute = lang


def prog1():
    print('\n === PROG 1 === \n')
    execute('''
    mem var iters 0;
    mem var k 2;
    point test; 
        p:test work ife 0 set 1;
        p:test op * $k;
        p:test print !;
        p:test out n;
    print Input iterations number:~;
    in num; mem edit iters !; >;
    out n; print $iters; out n;
    work repeat $iters goto test;
    load 10 test data for output;
    print data in 10 from~; print ?; print :~; print %10;
    ''')




def prog2():
    print('\n === PROG 2 === \n')
    execute('''
    # configs and systems;
    point input; p:input jump 1234;
    point basic; p:basic jump 0;
    point temp; p:temp jump 8421;
    
    point main;
        p:main goto temp;
        p:main copy 1235;
        p:main goto input;
        p:main op * >;
        p:main out .;
    # program;
    goto temp; print Input num 1:~; in num; goto basic;
    goto input; print Input num 2:~; in num; goto basic;
    goto main;
    >;
    ''')



def find_letter(string, symbol):
    # fast findin letter
    # example "prod" code architecture
    max = len(string) + 3
    code = f'''
# systems ------;       mem var max {max}; rlim 500; malloc $max; jump 0; mem var split_in -1;
# load input ---;       {'\n'.join([f'set {i}; >;' for i in string])}{f'prev {len(string)};'}
# search block;

point go; 
point find;
point plus;

ps:find;
    mem var temp_pos ?;
    jump {max-1};
    set $split_in;
    work ife -1 mem edit split_in $now;
    jump $temp_pos;
pe:find;

ps:go; 
    mem var now ?;
    load  $max $split_in;
    work ifins {symbol} goto find; 
    >;
pe:go;

# return;
work repeat {len(string)} goto go; set $split_in; return;
'''
    return execute(code)








'''
Main rules of clean Bytex code and best practices 
for code-generation -
- Comentaries
- Code sections
- Optimal realization
- Full setting lang while begin
- Adaptive parameters and arch
Good example - find.
'''

if __name__ == '__main__':
    tests_count = 100
    times_bytex = []
    for i in range(tests_count): 
        time_s = time.time()
        print('iteration: ', i, ', res:', find_letter('1234567890123456789012345678901234567890', str(i)))
        time_e = time.time()
        times_bytex.append(time_e-time_s)
    print('bytex:', sum(times_bytex) / len(times_bytex) )
    times_vanila = []
    for i in range(tests_count): 
        time_s = time.time()
        print(i, '1234567890123456789012345678901234567890'.find(str(i)))
        time_e = time.time()
        times_vanila.append(time_e-time_s)
    print('vanila:', sum(times_vanila) / len(times_vanila) )